package parsing;

import static parsing.Scanner.TokenType.*;

public class Parser {

    ////////////////////////////////////////////////////////////////////////
    // Expression Interface                                               //
    ////////////////////////////////////////////////////////////////////////

    /** An abstract syntax tree representing an expression */
    public interface Expr {
        String infix();
        int eval();
    }

    ////////////////////////////////////////////////////////////////////////
    // Expression Implementations                                         //
    ////////////////////////////////////////////////////////////////////////

    // There's a lot of boilerplate here, but most of the code
    // "does the obvious thing"; the specifications are given by the
    // Expr interface.
    //
    // Therefore, we've written the code compactly.

    /** An expression containing only an integer. */
    public static class Int implements Expr {
        private final int n;
        public Int(int n) { this.n = n; }

        public String infix() { return " " + this.n + " "; }
        public int eval() { return n; }
    }

    /** The expression -e */
    public static class Negation implements Expr {
        private final Expr e;
        public Negation(Expr e) { this.e = e; }

        public String infix() { return "(-" + e.infix() + ")"; }
        public int eval() { return -e.eval(); }
    }

    /** Common helper class for expressions with two subexpressions (+,*,/). */
    public static abstract class BinaryExpr implements Expr {
        protected final Expr left, right;
        public BinaryExpr(Expr l, Expr r) { this.left = l; this.right = r; }

        protected abstract String op();
        public String infix() {
            return "(" + left.infix() + op() + right.infix() + ")";
        }
    }

    /** The expression e1 + e2 */
    public static class Sum extends BinaryExpr {
        public Sum(Expr l, Expr r) { super(l,r); }
        public @Override String op() { return "+"; }
        public @Override int eval() { return left.eval() + right.eval(); }
    }
    
    /** The expression e1 * e2 */
    public static class Product extends BinaryExpr {
        public Product(Expr l, Expr r) { super(l,r); }
        public @Override String op() { return "*"; }
        public @Override int eval() { return left.eval() * right.eval(); }
    }

    /** The expression e1 / e2 */
    public static class Quotient extends BinaryExpr {
        public Quotient(Expr l, Expr r) { super(l,r); }
        public @Override String op() { return "/"; }
        public @Override int eval() { return left.eval() / right.eval(); }
    }

    ////////////////////////////////////////////////////////////////////////
    // Parsing                                                            //
    ////////////////////////////////////////////////////////////////////////

    // Simple grammar:
    //
    //    Expr → Integer   
    //         |   ( Expr + Expr )

    public static Expr parseSimple(String s) {
        return parseSimple(new Scanner(s));
    }

    //    Expr → Integer   |   ( Expr + Expr )
    public static Expr parseSimple(Scanner s) {
		switch(s.peek()) {
		case INT:
			return new Int(s.consumeInt());
		case LPAREN:
			s.consume(LPAREN);
			Expr e1 = parseSimple(s);
			s.consume(PLUS);
			Expr e2 = parseSimple(s);
			s.consume(RPAREN);
			return new Sum(e1, e2);
		default:
			throw new IllegalArgumentException();
		}
    }

    // Grammar:
    //
    //    Expr   → Term1   [+|-] Term2   [+|-] Term3   [+|-] ...
    //    Term   → Factor1 [*|/] Factor2 [*|/] Factor3 [*|/] ...
    //    Factor → Integer |  - Factor |   ( Expr )

    public static Expr parseExpr(String s) {
        return parseExpr(new Scanner(s));
    }

    //    Expr   → Term1   [+|-] Term2   [+|-] Term3   [+|-] ...
    public static Expr parseExpr(Scanner s) {
    	// create an expression e that will contain the running sum so far;
    	// e <- the first Term encountered
    	Expr e= parseTerm(s);
    	
       	
    	while (true) { // until we run out of terms:
    		// e <- (e <op> <term>) according to the operator and term just parsed
    		if (s.peek() == PLUS) {
    			s.consume(PLUS);
    			e= new Sum(e, parseTerm(s));
    		} else if (s.peek() == MINUS) {
    			s.consume(MINUS);
    			e= new Sum(e, new Negation(parseTerm(s)));
    		} else {
    			// if the next token is neither plus nor minus, we're done!
    			return e;
    		}
    	}
    }
    
    //    Term   → Factor1 [*|/] Factor2 [*|/] Factor3 [*|/] ...
    public static Expr parseTerm(Scanner s) {
      Expr e = parseFactor(s);
      while (true) {
        if (s.peek() == TIMES) {
          s.consume(TIMES);
          e = new Product(e, parseFactor(s));
        } else if (s.peek() == DIVIDE) {
          s.consume(DIVIDE);
          e = new Quotient(e, parseFactor(s));
        } else {
          return e;
        }
      }
    }
    
    //    Factor → Integer |  - Factor |   ( Expr )
    public static Expr parseFactor(Scanner s) {
      switch (s.peek()) {
        case INT:
          return new Int(s.consumeInt());
        case MINUS:
          s.consume(MINUS);
          return new Negation(parseFactor(s));
        case LPAREN:
          s.consume(LPAREN);
          Expr e = parseExpr(s);
          s.consume(RPAREN);
          return e;
        default:
          throw new IllegalArgumentException();
      }
    }

    public static void testSimple() {
      Expr result = parseSimple("((10 + 4) + 8)");
      System.out.println("Test for simple Expr (sums only):");
      System.out.println(result.infix());
      System.out.println(result.eval());
    }
    public static void testExpr() {
    	Expr result = parseExpr("((1 + 2) * 4 + 17 - (6+2))/4");
    	System.out.println("Test for general Expr:");
    	System.out.println(result.infix());
    	System.out.println(result.eval());
    }
    public static void main(String[] args) {
        Parser.testSimple();
    	Parser.testExpr();

    }
}