package parsing;

/**
 * A Scanner reads tokens one at a time from an input source.  This scanner
 * reads characters from a String, and returns the tokens defined by TokenType
 * below.
 */
public class Scanner {

    public enum TokenType {
        /** a sequence of digits */
        INT,

        /** the symbols '(', ')', '+', '-', '*', and '/' */
        LPAREN, RPAREN, PLUS, MINUS, TIMES, DIVIDE,

        /** indicates that all tokens have been consumed. */
        END
    }


    ////////////////////////////////////////////////////////////////////////
    // Public methods                                                     //
    ////////////////////////////////////////////////////////////////////////

    /**
     * A Scanner that reads from the given input string.
     * @throws IllegalArgumentException if the input is malformed.  Note that
     *                                  the exception may not be thrown right
     *                                  away if the beginning of the input is
     *                                  well formed.
     */
    public Scanner(String input) throws IllegalArgumentException {
        this.input   = input;
        this.n       = input.length();
        this.nextPos = 0;

        advance();
    }

    /** return the type of the next token */
    public TokenType peek() {
        return this.peek;
    }

    /**
     * check that the next token is of type t, and consume it if so.
     * @throws IllegalArgumentException if the next token is not of type t
     */
    public void consume(TokenType t) throws IllegalArgumentException {
        if (t != this.peek)
            throw new IllegalArgumentException ("expected " + t + " but got " + this.peek);

        advance();
    }

    /**
     * consume and return an integer from the input stream.
     * @throws IllegalArgumentException if the next token is not an int.
     */
    public int consumeInt() throws IllegalArgumentException {
        if (this.peek != TokenType.INT)
            throw new IllegalArgumentException ("expected INT but got " + this.peek);

        int result = Integer.parseInt(input.substring(pos,nextPos));
        advance();
        return result;
    }

    /** return the unconsumed portion of the input. */
    public String remainder() {
        return this.input.substring(nextPos);
    }

    ////////////////////////////////////////////////////////////////////////
    // Implementation                                                     //
    ////////////////////////////////////////////////////////////////////////

    /**
     * class invariants:
     *   n = input.length
     *
     *          |0          |pos                        |nextPos         |n
     *   input: [ processed | single token of type peek | unprocessed    ]
     */
    private final String input;
    private final int    n;

    private int       pos;
    private int       nextPos;
    private TokenType peek;

    /**
     * update pos, nextPos, and peek so that input[pos...nextPos-1] contains
     * the next token.
     *
     *  @throws IllegalArgumentException if the input is malformed.
     */
    private void advance() {
        // skip over whitespace
        while (nextPos < n && Character.isWhitespace(input.charAt(nextPos)))
            nextPos += 1;

        pos = nextPos;

        // handle EOF
        if (n == pos) {
            peek = TokenType.END;
            return;
        }

        switch (input.charAt(pos)) {
        // single character tokens:
        case '+': peek = TokenType.PLUS;   nextPos = pos+1; return;
        case '-': peek = TokenType.MINUS;  nextPos = pos+1; return;
        case '*': peek = TokenType.TIMES;  nextPos = pos+1; return;
        case '/': peek = TokenType.DIVIDE; nextPos = pos+1; return;
        case '(': peek = TokenType.LPAREN; nextPos = pos+1; return;
        case ')': peek = TokenType.RPAREN; nextPos = pos+1; return;

        // integers
        default:
            if (!Character.isDigit(input.charAt(pos)))
                throw new IllegalArgumentException("invalid character " + input.charAt(pos));

            peek = TokenType.INT;
            while (nextPos < n && Character.isDigit(input.charAt(nextPos)))
                nextPos += 1;
        }
    }
}